@@ -6,6 +6,7 @@ from configparser import ConfigParser
 from typing import List, TextIO
 
 import attr
+from mgedb import MGEdb
 
 
 @attr.s(auto_attribs=True, frozen=True, slots=True)
@@ -20,6 +21,7 @@ class ExecutionContext:
     blastn_path: str
     makeblastdb_path: str
     config: ConfigParser
+    mge_db: MGEdb
     fq_files: List[str]
     base_wd_dir: str
     timestamp: datetime.datetime = datetime.datetime.now()
