@@ -1,49 +1,25 @@
 """Functions for finding and validating MGEs. """
 import logging
 import math
+from collections import defaultdict
 from configparser import ConfigParser
-from itertools import groupby
+from itertools import chain, groupby
 from typing import Dict, Iterator, List, Optional, Tuple, Union
 
 import attr
-from mypy_extensions import TypedDict
-
 from mgedb import MGEdb, Sequence, Sequences
 from mgedb.db import MgeType
+from mypy_extensions import TypedDict
 
 from .context import ExecutionContext
 from .errors import CoordOutOfBoundsError
 from .io import ContigSequences
-from .result import Alignment, MgeFinderResult, MgeResult, PredictionEvidence, TemplateSequence
+from .result import (Alignment, MgeFinderResult, MgeResult, PredictionEvidence,
+                     TemplateSequence)
 from .tools import BlastHit, BlastHsp, iter_blast_hits
 
 LOG = logging.getLogger(__name__)
 
-# Type declaration
-HSP_BIN = Tuple[BlastHsp, ...]
-HSP_BINS = Tuple[HSP_BIN, ...]
-ELEM_COORD = Tuple[int, int]
-ALIGNMENTS = Tuple[Alignment, ...]
-
-
-@attr.s(auto_attribs=True, frozen=True, slots=True)
-class ElementHit:
-    """Blast hit."""
-
-    query_name: str
-    query_length: int
-    subject_name: str
-    tot_aln_cov: float
-    avg_seq_id: float
-    n_gaps: int
-    n_subs: int
-    element_len: int
-    subject_len: int
-    depth: Optional[float]
-    e_value: float
-    cigar: str
-    hsps: Tuple[BlastHsp, ...]
-
 
 def _is_valid_hit(hit: ElementHit, cfg: ConfigParser) -> bool:
     """Validate good hits."""
@@ -66,7 +42,7 @@ def _is_valid_hit(hit: ElementHit, cfg: ConfigParser) -> bool:
     return True
 
 
-def _calc_tot_aln_cov(group: HSP_BIN, hit: BlastHit) -> Union[int, float]:
+def _calc_tot_aln_cov(group: HspBin, hit: BlastHit) -> Union[int, float]:
     """Calculate total alignement coverage of group of hsps."""
     tot_len = sum(_hsp_len(hsp) for hsp in group)
     num_gaps = sum(int(hsp.num_gaps) for hsp in group)
@@ -74,7 +50,7 @@ def _calc_tot_aln_cov(group: HSP_BIN, hit: BlastHit) -> Union[int, float]:
     return (tot_len - num_gaps) / subject_len
 
 
-def _avg_seq_ident(group: HSP_BIN, hit: BlastHit) -> float:
+def _avg_seq_ident(group: HspBin, hit: BlastHit) -> float:
     """Calculate total alignement coverage of group of hsps."""
     ident_nt = sum(float(hsp.identity) * _hsp_len(hsp) for hsp in group)
     return ident_nt / sum(_hsp_len(hsp) for hsp in group)
@@ -86,11 +62,28 @@ def _hsp_len(hsp: BlastHsp) -> int:
     return r_end - r_start + 1
 
 
-def _get_elem_coord(mge_hit: ElementHit) -> ELEM_COORD:
+def _get_element_coord(mge: ElementHit) -> ElementCoordinate:
     """Get start and end coordinates from hit."""
-    start = min(mge_hit.hsps, key=lambda x: x.query_start).query_start
-    end = max(mge_hit.hsps, key=lambda x: x.query_end).query_end
-    return start, end
+    if isinstance(mge, MgeResult):
+        return mge.start, mge.end
+    elif isinstance(mge, ElementHit):
+        start = min(mge.hsps, key=lambda x: x.query_start).query_start
+        end = max(mge.hsps, key=lambda x: x.query_end).query_end
+        return start, end
+    else:
+        raise ValueError
+
+
+def _get_alignment_coords_elem(mge: ElementHit) -> ElementCoordinate:
+    """Get start and end coordinates from hit."""
+    if isinstance(mge, MgeResult):
+        return mge.start, mge.end
+    elif isinstance(mge, ElementHit):
+        start = min(mge.hsps, key=lambda x: x.query_start).query_start
+        end = max(mge.hsps, key=lambda x: x.query_end).query_end
+        return start, end
+    else:
+        raise ValueError
 
 
 def _get_possible_elem_coord(hsps: List[BlastHsp],
@@ -116,7 +109,7 @@ def _get_possible_elem_coord(hsps: List[BlastHsp],
         int(seg_start + (subject_length * padding)))
 
 
-def _group_hsps(hit: BlastHit) -> HSP_BINS:
+def _group_hsps(hit: BlastHit) -> HspBinS:
     """Group HSPs that are likely part of same MGE.
 
     HSP considered part of mge if its wihtin len(template mge) - len(ref hsp)
@@ -179,7 +172,7 @@ def _bin_hits(hits: Iterator[ElementHit]):
     """
     bins = [[next(hits)]]
     for hit in hits:
-        elem_coord = _get_elem_coord(hit)
+        elem_coord = _get_element_coord(hit)
         bin_idx = None
         for idx, curr_bin in enumerate(bins):
             reprs = curr_bin[0]  # get element representative for group
@@ -196,7 +189,7 @@ def _bin_hits(hits: Iterator[ElementHit]):
     return bins
 
 
-def _is_contained(first: ELEM_COORD, second: ELEM_COORD) -> bool:
+def _is_contained(first: ElementCoordinate, second: ElementCoordinate) -> bool:
     """Is first element nested in second element.
 
     Each element consists of a list with paired start, end values.
@@ -206,8 +199,8 @@ def _is_contained(first: ELEM_COORD, second: ELEM_COORD) -> bool:
     return s_start <= f_start and f_end <= s_end
 
 
-def _is_overlapping(first: ELEM_COORD,
-                    second: ELEM_COORD,
+def _is_overlapping(first: ElementCoordinate,
+                    second: ElementCoordinate,
                     allowed_overlap: int = 5) -> bool:
     """Check if first gene is overlapping second."""
     if allowed_overlap < 0:
@@ -223,7 +216,7 @@ def _is_overlapping(first: ELEM_COORD,
 
 def _pad_mge_coord(elem: ElementHit, padding: float) -> Tuple[int, int]:
     """Get range around element padded with X%."""
-    start, end = _get_elem_coord(elem)
+    start, end = _get_element_coord(elem)
     length = (end - start) * padding
     nstart = start - (length / 2) if start - (length / 2) >= 0 else 0
     return round(nstart), round(end + (length / 2))
@@ -235,7 +228,7 @@ def _get_mge_hit(blast_hit: ElementHit) -> Dict[str, str]:
     return {'name': name, 'seq_no': allele, 'accession': accnr}
 
 
-def _join_cigar(hsps: HSP_BIN) -> str:
+def _join_cigar(hsps: HspBin) -> str:
     """Join hsp alignment cigar strings.
 
     Unaligned regions are denoted with N.
@@ -252,9 +245,35 @@ def _join_cigar(hsps: HSP_BIN) -> str:
     return cigar
 
 
-def _identify_mges(ctx: ExecutionContext, blast_result: str,
+def _has_preserved_ir(mge_hit, db):
+    """Check if MGE has preserved IR."""
+    if len(mge_hit.hsps) > 1:
+        first_hit, *_, last_hit =  sorted(mge_hit.hsps, key=lambda x: x.subject_start)
+    else:
+        first_hit = mge_hit.hsps[0]
+        last_hit = mge_hit.hsps[0]
+
+    min_segment_length = 60
+    min_ir_aln_length = 20
+    # get expected MGE length from MGEdb
+    mge_info = _get_mge_hit(mge_hit)
+    expected_len = len(db.records[mge_info['name']].sequences[int(mge_info['seq_no']) - 1])
+    preserved_irl = not any([min_ir_aln_length < first_hit.subject_start,  # check if irl is preserved
+                             len(first_hit) < min_segment_length])  # assert lenght of hsp
+    preserved_irr = not any([min_ir_aln_length < (expected_len - last_hit.subject_end),  # check if irr is preserved
+                             len(last_hit) < min_segment_length])  # assert lenght of hsp
+    return any([preserved_irl, preserved_irr])
+
+
+def _is_type(mge: ElementHit, mge_type: MgeType, db: MGEdb):
+    """Check if MGE is of TYPE."""
+    mge_name = _get_mge_hit(mge)['name']
+    return db.records[mge_name].type == mge_type
+
+
+def _identify_mges(ctx: ExecutionContext, raw_mges,
                    contigs: ContigSequences,
-                   db: MGEdb) -> Tuple[MgeFinderResult, Sequences, ALIGNMENTS]:
+                   db: MGEdb) -> Tuple[MgeFinderResult, Sequences, Alignments]:
     """Find valid mobile element from blast reult and annotate output.
 
     HSPs for each hit is combined and the total alignment is summarized. The quality of the total
@@ -264,34 +283,14 @@ def _identify_mges(ctx: ExecutionContext, blast_result: str,
 
     """
     cfg = ctx.config  # read config
-    valid_mges = []
-    for hit in iter_blast_hits(blast_result):  # type: BlastHit
-        for el_hsps in _group_hsps(hit):
-            # TODO find better way of combining e-value.
-            # summarize hit quality metrics
-            element = ElementHit(tot_aln_cov=_calc_tot_aln_cov(el_hsps, hit),
-                                 avg_seq_id=_avg_seq_ident(el_hsps, hit),
-                                 n_gaps=sum(h.num_gaps for h in el_hsps),
-                                 n_subs=sum(h.num_subs for h in el_hsps),
-                                 element_len=sum(_hsp_len(g) for g in el_hsps),
-                                 subject_len=hit.subject_len,
-                                 depth=hit.depth,
-                                 query_name=hit.query_name,
-                                 query_length=hit.query_len,
-                                 subject_name=hit.subject_name,
-                                 e_value=min(h.e_value for h in el_hsps),
-                                 cigar=_join_cigar(el_hsps),
-                                 hsps=el_hsps)
-
-            if _is_valid_hit(element, cfg):
-                valid_mges.append(element)
-
+    # get mge records
+    valid_mges = (mge for mge in raw_mges if _is_valid_hit(mge, cfg))
     mge_records = db.records  # read records to memeory
 
     mge_sequences: List[Sequence] = []
     mge_headers = []  # store fasta headers of selected seqs
-    mges = []
-    alignments = []
+    mges: List[MgeResult] = []
+    alignments: List[Alignment] = []
     mge_indexer = 1
     for _, hits in groupby(valid_mges, key=lambda x: x.query_name):
         for hit_bin in _bin_hits(hits):
@@ -301,7 +300,7 @@ def _identify_mges(ctx: ExecutionContext, blast_result: str,
             header_info = _get_mge_hit(best_hit)
             entry = mge_records[header_info['name']]
 
-            start, end = _get_elem_coord(best_hit)
+            start, end = _get_element_coord(best_hit)
             mge_seq = contigs.sub_sequence(start - 1, end, 1,
                                            best_hit.query_name)
 
@@ -377,15 +376,38 @@ def _identify_mges(ctx: ExecutionContext, blast_result: str,
     return tuple(mges), tuple(mge_sequences), tuple(alignments)
 
 
-def compose_cn_mge_id(mge_cnt, first_is, second_is):
+def compose_cn_mge_id(mge_cnt, *is_elements):
     """Build mge_id for composite transposons."""
-    return f'{mge_cnt}_{first_is.mge_id}_{second_is.mge_id}'
+    full_is = sorted((e for e in is_elements if isinstance(e, MgeResult)),
+                     key=lambda m: m.mge_id)
+    if len(full_is) == 2:
+        first_is, second_is = full_is
+        return f'{mge_cnt}_{first_is.mge_id}_{second_is.mge_id}'
+    return f'{mge_cnt}_{full_is[0].mge_id}'
+
+
+def _remove_overlapping_raw_mges(raw_mges, valid_mges):
+    """Remove overlapping raw mges."""
+    for r in raw_mges:
+        r_start = min(r.hsps, key=lambda m: m.query_start).query_start
+        r_end = min(r.hsps, key=lambda m: m.query_end).query_end
+        if not any(_is_overlapping((r_start, r_end), (m.start, m.end)) for m in valid_mges):
+            yield r
+
+
+def _get_putative_cn_element(*elements):
+    """Get metrics for putative composite transposon element."""
+    first_start = min(min(map(_get_element_coord, elements)))
+    second_end = max(max(map(_get_element_coord, elements)))
+    return second_end - first_start + 1
 
 
 def _infer_putative_composite_tn(
         ctx: ExecutionContext,
         mges: MgeFinderResult,
+        raw_mges: List[ElementHit, ],
         contigs: ContigSequences,
+        db: MGEdb,
 ) -> Tuple[MgeFinderResult, Sequences]:
     """Infer putative composite transposons from prediction result.
 
@@ -395,6 +417,14 @@ def _infer_putative_composite_tn(
     List of MgeRecords file.
     """
     cfg = ctx.config
+    # get raw elements that are insertion sequences and has preserev IR
+    # to reduce the number of hits to bin and iterate over
+    is_w_ir_per_contig = defaultdict(lambda: defaultdict(list))
+    for raw_mge in raw_mges:
+        mge_name = raw_mge.subject_name.split('|')[0]
+        if all((_has_preserved_ir(raw_mge, db), _is_type(raw_mge, MgeType.INSERTION_SEQUENCE, db))):
+            is_w_ir_per_contig[raw_mge.query_name][mge_name].append(raw_mge)
+
     RawEntry = TypedDict(
         'RawEntry', {
             'mge_id': int,
@@ -402,35 +432,64 @@ def _infer_putative_composite_tn(
             '5p_flank': MgeResult,
             '3p_flank': MgeResult,
             'contig': str,
-            'sequence': None
+            'sequence': None,
         })
-    raw_ctn: Dict[str, RawEntry] = {}
+    raw_ctn = defaultdict(list)
     # find potential composite transposon
-    for contig, vals in groupby(mges, key=lambda x: x.contig):
-        for _, identical_mges in groupby(vals, key=lambda x: x.name):
+    for contig, mges_on_contig in groupby(mges, key=lambda x: x.contig):
+        for _, identical_mges in groupby(mges_on_contig, key=lambda x: x.name):
             gr_mges: List[MgeResult] = list(identical_mges)  # allow for repeated iteration
             mge_id = int(max(gr_mges, key=lambda x: int(x.mge_id)).mge_id) + 1
-            ins_seq = sorted(
-                [mge for mge in gr_mges if mge.type == MgeType.INSERTION_SEQUENCE],
-                key=lambda x: x.start)
-            if len(ins_seq) > 1:
-                # check length for each insertion sequence pair
-                for i in range(len(ins_seq) - 1):
-                    first_is, second_is = ins_seq[
-                        i:i + 2]  # type: Tuple[MgeResult, MgeResult]
-                    segment_len = second_is.end - first_is.start + 1
-                    if segment_len < cfg.getint('composite_transposon', 'max_len'):
-                        # composite transposon
-                        raw_ctn[contig] = {
-                            'mge_id': compose_cn_mge_id(mge_id, first_is,
-                                                        second_is),
-                            'length': segment_len,
-                            '5p_flank': first_is,
-                            '3p_flank': second_is,
-                            'contig': contig,
-                            'sequence': None
-                        }
-                        mge_id += 1
+            for mge_name, valid_mges in groupby(gr_mges, key=lambda m: m.name):
+                valid_ins_seq = sorted(valid_mges, key=lambda x: x.start)
+                # skip mges that are not insertion sequences
+                if valid_ins_seq[0].type != MgeType.INSERTION_SEQUENCE:
+                    continue
+                # get relevant raw mges and remove elements that are overlapping with valid mges
+                non_overlapping_is = _remove_overlapping_raw_mges(
+                    is_w_ir_per_contig[contig][mge_name],
+                    valid_ins_seq)
+                merged_ins_seq = sorted(chain(valid_ins_seq, non_overlapping_is),
+                                        key=lambda m: m.start if isinstance(m, MgeResult) else min(m.hsps, key=lambda h: h.query_start).query_start)
+                # find IS segments that are within the range for being
+                # predicted as a composite transposon
+                for i in range(len(merged_ins_seq)):
+                    reference_element = merged_ins_seq[i]
+                    if not isinstance(reference_element, MgeResult):
+                        continue
+                    # look for mge within threshold length downstream of reference element
+                    if i - 1 >= 0:
+                        first_element = merged_ins_seq[i - 1]
+                        element_len = _get_putative_cn_element(reference_element, first_element)
+                        if element_len < cfg.getint('composite_transposon', 'max_len'):
+                            # composite transposon
+                            comptn_entry: RawEntry = {'mge_id': compose_cn_mge_id(mge_id, first_element, reference_element),
+                                                      'length': element_len,
+                                                      '5p_flank': first_element,
+                                                      '3p_flank': reference_element,
+                                                      'contig': contig,
+                                                      'sequence': None}
+                            raw_ctn[contig].append(comptn_entry)
+                            mge_id += 1
+                    # look for mge within threshold length upstream of reference element
+                    if i != len(merged_ins_seq) - 1:  # skip if last element
+                        second_element = merged_ins_seq[i + 1]
+                        # if another full MGE is located upstream of the reference MGE skip it
+                        # when predicting composite transposons as it will generate duplicate
+                        # predictions.
+                        if isinstance(second_element, MgeResult):
+                            continue
+                        element_len = _get_putative_cn_element(reference_element, merged_ins_seq[i + 1])
+                        if element_len < cfg.getint('composite_transposon', 'max_len'):
+                            # composite transposon
+                            comptn_entry: RawEntry = {'mge_id': compose_cn_mge_id(mge_id, reference_element, second_element),
+                                                      'length': element_len,
+                                                      '5p_flank': reference_element,
+                                                      '3p_flank': second_element,
+                                                      'contig': contig,
+                                                      'sequence': None}
+                            raw_ctn[contig].append(comptn_entry)
+                            mge_id += 1
 
     putative_ctn = []
     ctn_sequences = set()
@@ -438,21 +497,19 @@ def _infer_putative_composite_tn(
     if raw_ctn:
         for fa in contigs:  # type: Sequence
             contig_name: str = fa.title
-            entry = raw_ctn.get(contig_name)
-            if entry:
-                start: int = entry['5p_flank'].start - 1
-                end: int = entry['3p_flank'].end
+            for entry in raw_ctn.get(contig_name, []):
+                start: int = _get_element_coord(entry['5p_flank'])[0] - 1
+                end: int = _get_element_coord(entry['3p_flank'])[1]
                 sequence = fa.seq[start:end]
                 entry['sequence'] = sequence  # get sequence from contig
-                assert len(sequence) == entry['length']
-
+                full_is = entry['5p_flank'] if isinstance(entry['5p_flank'], MgeResult) else entry['3p_flank']
+                cn_name = f'cn_{entry["length"]}_{full_is.name}'
                 mge_id = entry['mge_id']
-                cn_name = f'cn_{entry["length"]}_{entry["5p_flank"].name}'
                 header = f'{cn_name}|{mge_id}'
                 ctn_sequences.add(Sequence(title=header, seq=sequence))
                 # make new mge entry with putative composite transposon
                 putative_ctn.append(
-                    attr.evolve(entry['5p_flank'],
+                    attr.evolve(full_is,
                                 mge_id=mge_id,
                                 name=cn_name,
                                 synonyms=[],
@@ -460,8 +517,8 @@ def _infer_putative_composite_tn(
                                 evidence=PredictionEvidence.PUTATIVE,
                                 reference=[],
                                 alias=[],
-                                start=entry['5p_flank'].start,
-                                end=entry['3p_flank'].end,
+                                start=start,
+                                end=end,
                                 allele_seq_length=entry['length'],
                                 template_length=None,
                                 carried_motifs=[],
@@ -473,9 +530,31 @@ def predict_mges(ctx: ExecutionContext, blast_result: str,
                  contigs: ContigSequences,
                  db: MGEdb) -> Tuple[MgeFinderResult, Sequences, ALIGNMENTS]:
     """Predict mges with annotated sequence depth."""
+    # Group MGEs and format them
+    raw_mges: List[ElementHit, ] = []
+    for hit in iter_blast_hits(blast_result):  # type: BlastHit
+        for el_hsps in _group_hsps(hit):
+            # TODO find better way of combining e-value.
+            # summarize hit quality metrics
+            element = ElementHit(tot_aln_cov=_calc_tot_aln_cov(el_hsps, hit),
+                                 avg_seq_id=_avg_seq_ident(el_hsps, hit),
+                                 n_gaps=sum(h.num_gaps for h in el_hsps),
+                                 n_subs=sum(h.num_subs for h in el_hsps),
+                                 element_len=sum(_hsp_len(g) for g in el_hsps),
+                                 subject_len=hit.subject_len,
+                                 depth=hit.depth,
+                                 query_name=hit.query_name,
+                                 query_length=hit.query_len,
+                                 subject_name=hit.subject_name,
+                                 e_value=min(h.e_value for h in el_hsps),
+                                 cigar=_join_cigar(el_hsps),
+                                 hsps=el_hsps)
+            raw_mges.append(element)
+
+    # predict mges
     mges, mge_sequences, alignments = _identify_mges(
-        ctx, blast_result, contigs,
+        ctx, raw_mges, contigs,
         db)  # type: (MgeFinderResult, Sequences, ALIGNMENTS)
     put_ctn, ctn_sequences = _infer_putative_composite_tn(
-        ctx, mges, contigs)  # type: (MgeFinderResult, Sequences)
+        ctx, mges, raw_mges, contigs, db)  # type: (MgeFinderResult, Sequences)
     return mges + put_ctn, mge_sequences + ctn_sequences, alignments
